package cz.vutbr.feec.imageprocessing.op.featureExtraction;

import ij.ImagePlus;
import ij.process.ByteProcessor;
import ij.process.ImageProcessor;

import java.util.List;

import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.operator.ports.metadata.SimplePrecondition;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.parameter.UndefinedParameterError;

import cz.vutbr.feec.imageprocessing.imagej.imageplus.ImagePlusClrIOObject;
import cz.vutbr.feec.imageprocessing.imagej.imageplus.ImagePlusGSIOObject;

public class BIC extends Operator {

	private static final String PARAM_TOLERANCE = "tolerance";

	protected InputPort inImg = getInputPorts().createPort("color image plus");

	private OutputPort gsImOutput = getOutputPorts().createPort(
			"grayscale image plus");
	private OutputPort gsImOutputHist = getOutputPorts().createPort(
			"grayscale image plus Hist");

	public BIC(OperatorDescription description) {
		super(description);
		inImg.addPrecondition(new SimplePrecondition(inImg, new MetaData(
				ImagePlusClrIOObject.class)));
		getTransformer().addGenerationRule(gsImOutput,
				ImagePlusGSIOObject.class);
		getTransformer().addGenerationRule(gsImOutputHist,
				ImagePlusGSIOObject.class);
	}

	@Override
	public void doWork() throws OperatorException {

		ImagePlusClrIOObject img = inImg.getData();
		ImagePlus i = img.getImage();

		if (i.getProcessor().getPixel(0, 0) >= 0) {
			throw new UserError(this, 136, "Image is not color.");
		}

		int[][] rgbArray = i.getProcessor().getIntArray();
		int width = i.getWidth();
		int height = i.getHeight();

		ImageProcessor ipRed = new ByteProcessor(width, height);
		ImageProcessor ipGreen = new ByteProcessor(width, height);
		ImageProcessor ipBlue = new ByteProcessor(width, height);
		ImageProcessor ipResult = new ByteProcessor(width, height);

		extractChannels(rgbArray, ipRed, ipGreen, ipBlue);

		quantization(ipRed, 4);
		quantization(ipGreen, 4);
		quantization(ipBlue, 4);

		makeBorderPic(ipRed, ipGreen, ipBlue, ipResult);

		ImageProcessor ipHistogram = new ByteProcessor(width, height);
		histograVisualisation(ipRed, ipGreen, ipBlue, ipResult, ipHistogram);

		ImagePlus imp = new ImagePlus("", ipResult);
		ImagePlus imH = new ImagePlus("", ipHistogram);
		ImagePlusGSIOObject outImg = new ImagePlusGSIOObject(img.getFileName()
				+ "_BIC", imp);
		ImagePlusGSIOObject outImgH = new ImagePlusGSIOObject(img.getFileName()
				+ "_BIC_Hist", imH);
		outImg.setFitnessMask(img.getFitnessMask());
		outImgH.setFitnessMask(img.getFitnessMask());
		gsImOutput.deliver(outImg);
		gsImOutputHist.deliver(outImgH);
	}

	private void makeBorderPic(ImageProcessor ipRed, ImageProcessor ipGreen,
			ImageProcessor ipBlue, ImageProcessor ipResult)
			throws UndefinedParameterError {
		int tolerance = getParameterAsInt(PARAM_TOLERANCE);
		for (int x = 1; x < ipRed.getWidth(); x++) {
			for (int y = 1; y < ipRed.getHeight(); y++) {
				int t = 0;
				if (ipRed.getPixel(x, y) == ipRed.getPixel(x + 1, y))
					t++;
				if (ipRed.getPixel(x, y) == ipRed.getPixel(x - 1, y))
					t++;
				if (ipRed.getPixel(x, y) == ipRed.getPixel(x, y + 1))
					t++;
				if (ipRed.getPixel(x, y) == ipRed.getPixel(x, y - 1))
					t++;
				if (ipGreen.getPixel(x, y) == ipGreen.getPixel(x + 1, y))
					t++;
				if (ipGreen.getPixel(x, y) == ipGreen.getPixel(x - 1, y))
					t++;
				if (ipGreen.getPixel(x, y) == ipGreen.getPixel(x, y + 1))
					t++;
				if (ipGreen.getPixel(x, y) == ipGreen.getPixel(x, y - 1))
					t++;
				if (ipBlue.getPixel(x, y) == ipBlue.getPixel(x + 1, y))
					t++;
				if (ipBlue.getPixel(x, y) == ipBlue.getPixel(x - 1, y))
					t++;
				if (ipBlue.getPixel(x, y) == ipBlue.getPixel(x, y + 1))
					t++;
				if (ipBlue.getPixel(x, y) == ipBlue.getPixel(x, y - 1))
					t++;
				if (t >= (12 - tolerance)) {
					ipResult.set(x, y, 255);
				}
			}
		}
	}

	private void histograVisualisation(ImageProcessor ipRed,
			ImageProcessor ipGreen, ImageProcessor ipBlue,
			ImageProcessor ipResult, ImageProcessor ipHistogram) {
		for (int x = 0; x < ipRed.getWidth(); x++) {
			for (int y = 0; y < ipRed.getHeight(); y++) {
				int red = ipRed.get(x, y) * 8;
				int green = ipGreen.get(x, y) * 32;
				int blue = ipBlue.get(x, y) * 16;
				int border = ipResult.get(x, y) / 255;

				ipHistogram.set(x, y, red + green + blue + border);

			}
		}

	}

	private void extractChannels(int[][] rgbArray, ImageProcessor ipRed,
			ImageProcessor ipGreen, ImageProcessor ipBlue) {
		int width = ipRed.getWidth();
		int height = ipRed.getHeight();
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				ipRed.set(x, y, getRed(rgbArray[x][y]));
				ipGreen.set(x, y, getGreen(rgbArray[x][y]));
				ipBlue.set(x, y, getBlue(rgbArray[x][y]));
			}
		}
	}

	private int getRed(int rgb) {
		return ((rgb >> 16) & 0xff);
	}

	private int getBlue(int rgb) {
		return (rgb & 0xff);
	}

	private int getGreen(int rgb) {
		return ((rgb >> 8) & 0xff);
	}

	private void quantization(ImageProcessor ip, double colorCount) {
		ip.add(-32);
		ip.multiply(colorCount / 256.0);
		// ip.multiply(256.0 / colorCount);
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types.add(new ParameterTypeInt(PARAM_TOLERANCE, "tolerance", 0, 12, 0));

		return types;
	}
}