/*
 *  Brno University of Technology
 *
 *  Copyright (C) 2009-2010 by Brno University of Technology and the contributors
 *
 *  Complete list of developers available at our web site:
 *
 *       http://spl.utko.feec.vutbr.cz
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses/.
 */
package cz.vutbr.feec.imageprocessing.op.featureExtraction;

import java.util.List;

import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeInt;

import cz.vutbr.feec.imageprocessing.imagej.imageplus.ImagePlusGSIOObject;

/**
 * The Class Block Difference of Inverse probabilities. Default size of block
 * should be 2x2.
 * 
 * @article{bb83842, AUTHOR = "Chun, Y.D. and Seo, S.Y. and Kim, N.C.", TITLE =
 *                   "Image retrieval using BDIP and BVLC moments", JOURNAL =
 *                   CirSysVideo, VOLUME = "13", YEAR = "2003", NUMBER = "9",
 *                   MONTH = "September", PAGES = "951-957", BIBSOURCE =
 *                   "http://www.visionbib.com/bibliography/applicat816.html#TT80264"
 *                   }
 */
public class BVLC extends BaseImageOperator {

	private static final String PARAM_REAL_WINDOW_SIZE = "window_size";

	public BVLC(OperatorDescription description) {
		super(description);
	}

	@Override
	public void doWork() throws OperatorException {
		// delete all previous records
		clearFeatures();

		int M = getParameterAsInt(PARAM_REAL_WINDOW_SIZE);
		ImagePlusGSIOObject imObj = inImg.getData();
		int height = imObj.getWindowHeight();
		int width = imObj.getWindowWidth();
		if (height < M || width < M) {
			throw new OperatorException("Window dimension must be smaller then window in Feature Extraction.");
		} else if (M == 0) {
			if (height == width) {
				M = imObj.getWindowHeight();
			} else {
				throw new OperatorException(
						"Window in Feature Extraction is not square. Please set value.");
			}
		} else if (M < 0) {
			throw new OperatorException("Bad window dimension.");
		}

		double[] values = new double[4];
		values[0] = sigma(M, 0, M / 2);
		values[1] = sigma(M, M / 2, 0);
		values[2] = sigma(M, M / 2, M / 2);
		values[3] = sigma(M, M / 2, -M / 2);

		double max = max(values);
		double min = min(values);
		double res = max - min;

		// add feature
		setFeature("BVLC", res);

		// deliver output
		outExSet.deliver(features);
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types.add(new ParameterTypeInt(PARAM_REAL_WINDOW_SIZE, "window size",
				2, 99999, 2));
		return types;
	}

	private double max(double[] values) {
		double max = values[0];
		for (int i = 1; i < values.length; i++) {
			if (max < values[i]) {
				max = values[i];
			}
		}
		return max;
	}

	private double min(double[] values) {
		double min = values[0];
		for (int i = 1; i < values.length; i++) {
			if (min > values[i]) {
				min = values[i];
			}
		}
		return min;
	}

	private double sigma(int M, int k, int l) throws OperatorException {
		double numberOfPixels = M * M;

		double mean = 0;
		double meanShifted = 0;
		double deviation = 0;
		double deviationShifted = 0;

		// mean value
		for (int y = 0; y < M; y++) {
			for (int x = 0; x < M; x++) {
				int pixIntensity = getCentralPixel(x, y);
				int pixIntensityShifted = getCentralPixel(k + x, l + y);

				// mean
				mean += pixIntensity;

				// mean shifted
				meanShifted += pixIntensityShifted;
			}
		}
		mean = mean / numberOfPixels;
		meanShifted = meanShifted / numberOfPixels;

		// Std dev
		for (int y = 0; y < M; y++) {
			for (int x = 0; x < M; x++) {
				int pixIntensity = getCentralPixel(x, y);
				;
				int pixIntensityShifted = getCentralPixel(k + x, l + y);

				// deviation
				deviation += Math.pow(pixIntensity - mean, 2);

				// deviation shifted
				deviationShifted += Math.pow(pixIntensityShifted - mean, 2);
			}
		}
		deviation = Math.sqrt(deviation / numberOfPixels);
		deviationShifted = Math.sqrt(deviationShifted / numberOfPixels);

		// local covariance
		double localCovariance = 0;
		for (int y = 0; y < M; y++) {
			for (int x = 0; x < M; x++) {
				int pixIntensity = getCentralPixel(x, y);
				int pixIntensityShifted = getCentralPixel(k + x, l + y);

				localCovariance += pixIntensity * pixIntensityShifted - mean
						* meanShifted;
			}
		}
		double res = 0;
		if (localCovariance != 0 && deviation != 0 && deviationShifted != 0) {
			res = (1.0 / numberOfPixels * localCovariance)
					/ (deviation * deviationShifted);
		}
		return res;
	}
}
