/*
 *  Brno University of Technology
 *
 *  Copyright (C) 2009-2010 by Brno University of Technology and the contributors
 *
 *  Complete list of developers available at our web site:
 *
 *       http://spl.utko.feec.vutbr.cz
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses/.
 */
package cz.vutbr.feec.imageprocessing.op.featureExtraction;

import ij.ImagePlus;

import java.util.List;

import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.GenerateNewMDRule;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.operator.ports.metadata.SimplePrecondition;
import com.rapidminer.parameter.ParameterType;

import cz.vutbr.feec.imageprocessing.imagej.imageplus.ImagePlusGSIOObject;

/**
 * This operator is using sliding window over an image and extracts features as
 * an example set. According to the label image evaluates the samples and
 * assigns the label for the features.
 * 
 */
public abstract class BaseImageOperator extends Operator {

	protected InputPort inImg = getInputPorts().createPort("grayscale image plus");
	protected OutputPort outExSet = getOutputPorts().createPort("feature");

	protected FeatureIOObject features = new FeatureIOObject();

	public BaseImageOperator(OperatorDescription description) {
		super(description);
		inImg.addPrecondition(new SimplePrecondition(inImg, new MetaData(
				ImagePlusGSIOObject.class)));
		getTransformer().addRule(
				new GenerateNewMDRule(outExSet, FeatureIOObject.class));
	}

	/**
	 * Clear features - deletes all records.
	 */
	public void clearFeatures() {
		features.clear();
	}

	/**
	 * This method copy data from imagePlus to window.
	 * 
	 * @param imagePlus
	 * @param window
	 * @param x
	 * @param y
	 */
	protected void copyDataToWindow(ImagePlus imagePlus, byte[][] window,
			int x, int y) {
		for (int ix = 0; ix < window.length; ix++) {
			for (int iy = 0; iy < window[0].length; iy++) {
				window[ix][iy] = (byte) imagePlus.getProcessor().get(x + ix,
						y + iy);
			}
		}
	}

	/**
	 * Gets the central pixel of the current window.
	 * 
	 * @param x
	 *            the shift in x-axis direction (positive values move right)
	 * @param y
	 *            the shift in y-axis direction (positive values move down)
	 * 
	 * @return the central pixel
	 * @throws OperatorException
	 */
	protected int getCentralPixel(int x, int y) throws OperatorException {
		ImagePlusGSIOObject img = inImg.getData();
		int absX = x + img.getWindowXPosition();
		int absY = y + img.getWindowYPosition();
		// check bounds
		if (x < -img.getWindowWidth() / 2
				|| x > img.getWindowWidth() / 2 + img.getWindowWidth() % 2 - 1
				|| y < -img.getWindowHeight() / 2
				|| y > img.getWindowHeight() / 2 + img.getWindowHeight() % 2
						- 1) 
		{
			throw new OperatorException("Out of Feature Extraction Window bound.");
		}
		try {
			// System.out.println(resX+" "+resY);
			return img.getImage().getProcessor().get(absX, absY);
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new UserError(this, 116,
					"Please check that window width and "
							+ "height of Feature Extractor operator is "
							+ "sufficient. Requested position of image was X:"
							+ absX + ", Y:" + absY
							+ " but allowed range is X:0-"
							+ img.getImage().getWidth() + " Y:0-"
							+ img.getImage().getHeight());
		}
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		// add params here
		return types;
	}

	protected boolean isInnerPixel(int x, int y) {
		ImagePlusGSIOObject img;
		boolean inner = false;
		try {
			img = inImg.getData();
			int resX = img.getWindowXPosition() + x;
			int resY = img.getWindowYPosition() + y;
			int width = img.getImage().getWidth();
			int height = img.getImage().getHeight();
			if (resX >= 0 && resX < width && resY >= 0 && resY < height) {
				inner = true;
			}

		} catch (OperatorException e) {
			e.printStackTrace();
		}
		return inner;
	}

	/**
	 * Sets the feature with name to a given value.
	 * 
	 * @param name
	 *            the name
	 * @param value
	 *            the value
	 */
	public void setFeature(String name, double value) {
		features.setFeature(getName() + "_" + name, value);
	}

	/**
	 * Sets the feature.
	 * 
	 * @param name
	 *            the name
	 * @param value
	 *            the value
	 * @param ontology
	 *            the ontology
	 */
	public void setFeature(String name, Object value, int ontology) {
		features.setFeature(getName() + "_" + name, value, ontology,
				FeatureType.ATTRIBUTE);
	}

	/**
	 * Sets the feature.
	 * 
	 * @param name
	 *            the name
	 * @param value
	 *            the value
	 * @param ontology
	 *            the ontology
	 * @param isSpecial
	 *            the is special
	 */
	public void setFeature(String name, Object value, int ontology,
			FeatureType type) {
		features.setFeature(getName() + "_" + name, value, ontology, type);
	}

}
