/*
 *  Brno University of Technology
 *
 *  Copyright (C) 2009-2010 by Brno University of Technology and the contributors
 *
 *  Complete list of developers available at our web site:
 *
 *       http://spl.utko.feec.vutbr.cz
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses/.
 */
package cz.vutbr.feec.imageprocessing.op.featureExtraction;

import java.util.List;
import java.util.Vector;

import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.parameter.ParameterTypeInt;

public class CirclePixelsExtractor extends BaseImageOperator {

	private static final String PARAM_RADIUS = "radius";
	private static final String PARAM_ONLY_SMUMMARY = "only_summary";

	public CirclePixelsExtractor(OperatorDescription description) {
		super(description);
	}

	@Override
	public void doWork() throws OperatorException {
		// delete all previous records
		clearFeatures();
		int radius = getParameterAsInt(PARAM_RADIUS);
		if(radius < 1){
			throw new OperatorException("Radius must be positive number.");
		}
		//bounds check
		if((!isInnerPixel(-radius, -radius)) || (!isInnerPixel(radius, radius))){
			setFeature("AVG", 0);
			setFeature("STDEV", 0);

			// deliver output
			outExSet.deliver(features);
			return;
		}
		int x, y, r2;
		r2 = radius * radius;

		Vector<Double> data = new Vector<Double>();

		data.add((double) getCentralPixel(0, radius));
		data.add((double) getCentralPixel(0, -radius));
		data.add((double) getCentralPixel(radius, 0));
		data.add((double) getCentralPixel(-radius, 0));

		y = radius;
		x = 1;
		y = (int) (Math.sqrt(r2 - 1) + 0.5);
		while (x < y) {
			if(!isInnerPixel(-x, -y) || !isInnerPixel(x, y)){
				setFeature("AVG", 0);
				setFeature("STDEV", 0);
				
				// deliver output
				outExSet.deliver(features);
				return;
			}
			data.add((double) getCentralPixel(x, y));
			data.add((double) getCentralPixel(x, -y));
			data.add((double) getCentralPixel(-x, y));
			data.add((double) getCentralPixel(-x, -y));
			data.add((double) getCentralPixel(y, x));
			data.add((double) getCentralPixel(y, -x));
			data.add((double) getCentralPixel(-y, x));
			data.add((double) getCentralPixel(-y, -x));

			x += 1;
			y = (int) (Math.sqrt(r2 - x * x) + 0.5);
		}
		if (x == y) {
			data.add((double) getCentralPixel(+x, +y));
			data.add((double) getCentralPixel(+x, -y));
			data.add((double) getCentralPixel(-x, +y));
			data.add((double) getCentralPixel(-x, -y));
		}
		boolean onlySummary = getParameterAsBoolean(PARAM_ONLY_SMUMMARY);
		// Points
		double sum = 0;
		for (int i = 0; i < data.size(); i++) {
			if (!onlySummary) {
				setFeature("point" + i, data.get(i));
			}
			sum += data.get(i);
		}

		// Average value
		double avg = sum / data.size();
		setFeature("AVG", avg);

		// Standard deviation
		double stdev = 0;
		for (int i = 0; i < data.size(); i++) {
			double tmp = avg - data.get(i);
			stdev += tmp * tmp;
		}
		setFeature("STDEV", Math.sqrt(stdev));

		// deliver output
		outExSet.deliver(features);
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types.add(new ParameterTypeInt(PARAM_RADIUS, "Radius of the circle", 1,
				99999, 2));
		types
				.add(new ParameterTypeBoolean(
						PARAM_ONLY_SMUMMARY,
						"Returns only mean and std. dev. of values at the given radius.",
						true));
		return types;
	}
}
