/*
 *  Brno University of Technology
 *
 *  Copyright (C) 2009-2010 by Brno University of Technology and the contributors
 *
 *  Complete list of developers available at our web site:
 *
 *       http://spl.utko.feec.vutbr.cz
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses/.
 */

/*
 * This operator is contructed co compute Line-Haar like feature - horizontal and vertical
 */
package cz.vutbr.feec.imageprocessing.op.featureExtraction;

import ij.ImagePlus;

import java.util.List;

import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.parameter.ParameterTypeCategory;

import cz.vutbr.feec.imageprocessing.imagej.imageplus.ImagePlusGSIOObject;

public class LineHaarLikeFeature extends BaseImageOperator
{
	public LineHaarLikeFeature(OperatorDescription description)
	{
		super(description);
		// TODO Auto-generated constructor stub
	}
	
	private final String PARAM_X_SIZE = "x_size";
	private final String PARAM_Y_SIZE = "y_size";
	private final String PARAM_X_POSITION = "x_position";
	private final String PARAM_Y_POSITION = "y_position";
	private final String[] PARAM_HAAR_ORIENTATION = new String[] { "horizontal" , "vertical"};
	private final String PARAM_HAAR = "haar_orientation";
	
	
	@Override
	public void doWork() throws OperatorException
	{
		// delete all previous records
		clearFeatures();
		ImagePlusGSIOObject img= inImg.getData();
		ImagePlus im = img.getImage();
		
		//loading variables
		int xPosition = getParameterAsInt(PARAM_X_POSITION);
		int yPosition = getParameterAsInt(PARAM_Y_POSITION);
		int xSize = getParameterAsInt(PARAM_X_SIZE);
		int ySize = getParameterAsInt(PARAM_Y_SIZE);
		boolean haar;
		//determination of feature direction
		if( getParameterAsInt(PARAM_HAAR)==0) haar=true;
		else haar=false;
		
		//loading image and window
		int xWin = img.getWindowXPosition();
		int yWin = img.getWindowYPosition();
		int wWin = img.getWindowWidth();
		int hWin = img.getWindowHeight();
		
		
		//checking the parameter values
		//feature size for horiontal feature
		if(haar == true && (xSize%3 != 0 || xSize == 0 || ySize == 0))
			throw new UserError(this, 116, "Invalid input of \"X size\" or \"Y size\"  for Horizontal Line Haar-like feature");
		//feature size for vertical feature
		if(haar == false && (ySize%3 != 0 || ySize == 0 || xSize == 0))
			throw new UserError(this, 116, "Invalid input of \"X size\" or \"Y size\"  for Vertical Line Haar-like feature");
		//feature in the window
		if((xSize + xPosition) > wWin || (ySize + yPosition) > hWin)
			throw new UserError(this, 116, "Feature is out of window");

		
		int black = 0;
		int white = 0;
		int diff = 0;
		boolean b = true;
		
		//Horizontal feature computation
		if (haar == true){
			for (int step = xWin+xPosition; step< xWin+xPosition+xSize; step+=xSize/3) {
				//switching between white and black pixels
				b = !b;
				for (int i = step; i < step+xSize/3; i++) {
					for (int j = yWin+yPosition; j< yWin+yPosition+ySize;j++) {
						if (b==true) white += im.getProcessor().getPixel(i, j);
						else black += im.getProcessor().getPixel(i, j);
					}
				}
			}
		}
		//Vertical feature computation
		else {
			for (int step = yWin+yPosition; step< yWin+yPosition+ySize; step+=ySize/3) {
				//switching between white and black pixels
				b = !b;
				for (int i = step; i < step+ySize/3; i++) {
					for (int j = xWin+xPosition; j< xWin+xPosition+xSize;j++) {
						if (b==true) white += im.getProcessor().getPixel(j, i);
						else black += im.getProcessor().getPixel(j, i);
					}
				}
			}
		}
		
		diff = white - black;
		setFeature("LineHaarFeature", diff);
		// deliver output
		outExSet.deliver(features);
	}
	
	@Override
	public List<ParameterType> getParameterTypes()
	{
		
		List<ParameterType> params =  super.getParameterTypes();
		params.add(new ParameterTypeInt(
				PARAM_X_SIZE, 
				"X-size of the Haar like feature.", 
				1, Integer.MAX_VALUE, 3, true));
		params.add(new ParameterTypeInt(
				PARAM_Y_SIZE, 
				"Y-size of the Haar like feature.", 
				1, Integer.MAX_VALUE, 1, true));
		params.add(new ParameterTypeInt(
				PARAM_X_POSITION, 
				"Starting X-position.", 
				0, Integer.MAX_VALUE, 0, true));
		params.add(new ParameterTypeInt(
				PARAM_Y_POSITION, 
				"Starting Y-position.", 
				0, Integer.MAX_VALUE, 0, true));
		params.add(new ParameterTypeCategory(
				PARAM_HAAR, 
				"Haar orientation", PARAM_HAAR_ORIENTATION, 0));
		
		return params;
	}
}
