/*
 *  Brno University of Technology
 *
 *  Copyright (C) 2009-2010 by Brno University of Technology and the contributors
 *
 *  Complete list of developers available at our web site:
 *
 *       http://spl.utko.feec.vutbr.cz
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses/.
 */
package cz.vutbr.feec.imageprocessing.op.featureExtraction;

import ij.ImagePlus;
import ij.process.ByteProcessor;
import ij.process.ImageProcessor;

import java.util.List;
import java.util.Vector;

import com.rapidminer.example.Attribute;
import com.rapidminer.operator.IOObject;
import com.rapidminer.operator.OperatorChain;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.InputPortExtender;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.OutputPortExtender;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.operator.ports.metadata.OneToManyPassThroughRule;
import com.rapidminer.operator.ports.metadata.SimplePrecondition;
import com.rapidminer.operator.ports.metadata.SubprocessTransformRule;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.parameter.ParameterTypeCategory;
import com.rapidminer.tools.RandomGenerator;

import cz.vutbr.feec.imageprocessing.imagej.imageplus.ImagePlusGSIOObject;

/**
 * This operator is using sliding window over an image and extracts features as
 * an example set. According to the label image evaluates the samples and
 * assigns the label for the features.
 * 
 */
public class WaveletDecompositionOperator extends OperatorChain {
	// vnejsi vstup (velky obrazek)
	private InputPort imgIn = getInputPorts().createPort("image");

	// vnitrni zdroj (okno z puvodniho obrazku)
	private OutputPortExtender innerOut = new OutputPortExtender("window",
			getSubprocess(0).getInnerSources());

	// vnitrni vystup (okno z puvodniho obrazku)
	// private InputPortExtender innerIn = new InputPortExtender("example set",
	// getSubprocess(0).getInnerSinks());
	protected InputPortExtender innerIn = new InputPortExtender("feature",
			getSubprocess(0).getInnerSinks(), new MetaData(
					FeatureIOObject.class), true);

	// celkovy vystup
	private OutputPort exsOut = getOutputPorts().createPort("feature");

	List<Attribute> specialAttributes = new Vector<Attribute>();

	private final static String[] PARAM_DEGREE_LIST = { "1", "2", "3" };

	private static final String PARAM_DEGREE = "degree";

	private static final String PARAM_RECURSIVE = "recursive";
	private static final String PARAM_PACKET = "packet";

	public WaveletDecompositionOperator(OperatorDescription description)
			throws OperatorException {
		super(description, "Executed process");
		innerIn.start();
		innerOut.start();

		imgIn.addPrecondition(new SimplePrecondition(imgIn, new MetaData(
				ImagePlusGSIOObject.class)));

		getTransformer()
				.addRule(
						new OneToManyPassThroughRule(imgIn, innerOut
								.getManagedPorts()));
		getTransformer().addGenerationRule(exsOut, FeatureIOObject.class);
		getTransformer().addRule(new SubprocessTransformRule(getSubprocess(0)));
		// innerIn.makePassThroughRule(exsOut);
	}

	@Override
	public void doWork() throws OperatorException {

		int seed = getParameterAsInt(RandomGenerator.PARAMETER_LOCAL_RANDOM_SEED);

		ImagePlusGSIOObject imageIO = imgIn.getData();
		int imgWidth = imageIO.getImage().getWidth();
		int imgHeight = imageIO.getImage().getHeight();

		int x = imageIO.getWindowXPosition();
		int y = imageIO.getWindowYPosition();
		int winWidth = imageIO.getWindowWidth();
		int winHeight = imageIO.getWindowHeight();

		FeatureIOObject featureResultSet = new FeatureIOObject();

		boolean recursive = getParameterAsBoolean(PARAM_RECURSIVE);
		int levels = getParameterAsInt(PARAM_DEGREE);

		// for each level
		for (int l = 1; l <= levels; l++) {
			// include also sub-decompositions or is current level
			if (recursive || levels == l) {

				// for each wavelet transformation part
				int boxes = (int) Math.pow(2, levels);
				for (int ix = 0; ix < boxes; ix++) {
					for (int iy = 0; iy < boxes; iy++) {

						// window
						ImageProcessor ip = new ByteProcessor(imgWidth,
								imgHeight);
						ImagePlus imp = new ImagePlus("Wavelet", ip);
						ImagePlusGSIOObject imageIOWavelet = new ImagePlusGSIOObject(
								"WaveletTranform", imp);

						// set position and window size in the wavelet image
						imageIOWavelet.setWindowXPosition(x);
						imageIOWavelet.setWindowYPosition(y);
						imageIOWavelet.setWindowWidth(winWidth);
						imageIOWavelet.setWindowHeight(winHeight);

						// TODO do wavelet tranform and fill in the
						// imageIOWavelet

						// set coordinates on wavelet window where the data are
						// to be
						// collected
						int wavLevel = -1;
						int wavXPos = -1;
						int wavYPos = -1;

						// evaluate all inner operators and collect resulting
						// values
						for (IOObject o : innerIn.getData(false)) {
							FeatureIOObject ef = (FeatureIOObject) o;

							// evaluate features
							final boolean NO_CLONE = false;
							innerOut.deliverToAll(imageIOWavelet, NO_CLONE);
							getSubprocess(0).execute();
							inApplyLoop();

							// collect results
							for (Feature feature : ef) {
								checkForStop();
								String waveletPositionInfo = wavLevel + "_"
										+ wavXPos + "_" + wavYPos + "_";
								featureResultSet.setFeature(waveletPositionInfo
										+ feature.getName(), (Double) feature
										.getValue());
							}
						}
					}
				}
			}
		}
		exsOut.deliver(featureResultSet);
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types.add(new ParameterTypeCategory(PARAM_DEGREE,
				"The degree of wavelet transform decomposition.",
				PARAM_DEGREE_LIST, 0));
		types.add(new ParameterTypeBoolean(PARAM_RECURSIVE,
				"Should be all the sub-levels included to generation.", false));
		types.add(new ParameterTypeBoolean(PARAM_PACKET,
				"Use wavelet packet transform", false));
		types.get(0).setExpert(false);
		return types;
	}
}
