/*
 *  Brno University of Technology
 *
 *  Copyright (C) 2009-2010 by Brno University of Technology and the contributors
 *
 *  Complete list of developers available at our web site:
 *
 *       http://spl.utko.feec.vutbr.cz
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses/.
 */
package cz.vutbr.feec.imageprocessing.op.featureExtraction.global;

import ij.ImagePlus;
import ij.measure.Measurements;
import ij.process.ImageProcessor;
import ij.process.ImageStatistics;

import java.util.List;

import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.operator.ports.metadata.SimplePrecondition;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.tools.Ontology;

import cz.vutbr.feec.imageprocessing.imagej.imageplus.ImagePlusGSIOObject;
import cz.vutbr.feec.imageprocessing.op.featureExtraction.FeatureIOObject;
import cz.vutbr.feec.imageprocessing.op.featureExtraction.FeatureType;

public class StatisticsOperator extends Operator implements Measurements {

	private static final String PARAM_INCLUDE_MEAN = "Mean";

	private static final String PARAM_INCLUDE_MEDIAN = "Median";
	private static final String PARAM_INCLUDE_KURTOSIS = "Kurtosis";
	private static final String PARAM_INCLUDE_STDEV = "Standard Deviation";
	private static final String PARAM_INCLUDE_PEAK = "Peak";
	private static final String PARAM_INCLUDE_MIGV = "Min Gray Value";
	private static final String PARAM_INCLUDE_MAGV = "Max Gray Value";
	private static final String PARAM_INCLUDE_MASSCENT = "Center of Mass";
	private static final String PARAM_INCLUDE_SKEWNESS = "Skewness";
	private static final String PARAM_INCLUDE_ARFRAC = "Area Fraction";
	private static final String PARAM_INCLUDE_EDGE = "Edginess";
	protected InputPort inImg = getInputPorts().createPort(
			"grayscale image plus");

	private OutputPort gsImOutput = getOutputPorts().createPort(
			"grayscale image plus");
	
	private OutputPort globaFeature = getOutputPorts().createPort(
	"global features");

	public StatisticsOperator(OperatorDescription description) {
		super(description);
		inImg.addPrecondition(new SimplePrecondition(inImg, new MetaData(
				ImagePlusGSIOObject.class)));
		getTransformer().addGenerationRule(gsImOutput,
				ImagePlusGSIOObject.class);
		getTransformer().addGenerationRule(globaFeature, FeatureIOObject.class);
	}

	@Override
	public void doWork() throws OperatorException {
		ImagePlusGSIOObject imObj = inImg.getData();
		ImagePlus imp = imObj.getImage();
		
		FeatureIOObject feature = new FeatureIOObject();
		
		if(imp.getProcessor().getPixel(0, 0)  < 0) {
			throw new UserError(this, 136, "Image is not grayscale.");
		}
		
		ImageStatistics stats = imp.getStatistics(AREA + AREA_FRACTION
				+ SKEWNESS + CENTER_OF_MASS + MIN_MAX + MODE + MEDIAN
				+ KURTOSIS + MEAN + STD_DEV);

		if (getParameterAsBoolean(PARAM_INCLUDE_MEAN)) {
			feature.setFeature("Mean", stats.mean, Ontology.REAL,
					FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_MEDIAN)) {
			feature.setFeature("Median", stats.median, Ontology.REAL,
					FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_ARFRAC)) {
			feature.setFeature("Area Fraction", stats.areaFraction,
					Ontology.REAL, FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_KURTOSIS)) {
			feature.setFeature("KURTOSIS", stats.kurtosis, Ontology.REAL,
					FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_MAGV)) {
			feature.setFeature("Max Gray Value", stats.max, Ontology.REAL,
					FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_MASSCENT)) {
			feature.setFeature("X Center of Mass", stats.xCenterOfMass,
					Ontology.REAL, FeatureType.ATTRIBUTE);
			feature.setFeature("Y Center of Mass", stats.yCenterOfMass,
					Ontology.REAL, FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_MIGV)) {
			feature.setFeature("Minimum Gray Value", stats.min, Ontology.REAL,
					FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_PEAK)) {
			feature.setFeature("Peak", stats.mode, Ontology.REAL,
					FeatureType.ATTRIBUTE);
			feature.setFeature("Peak Relative count", imp.getProcessor()
					.getHistogram()[stats.mode]
					/ stats.area, Ontology.REAL, FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_SKEWNESS)) {
			feature.setFeature("Skewness", stats.skewness, Ontology.REAL,
					FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_STDEV)) {
			feature.setFeature("Standard Deviation", stats.stdDev, Ontology.REAL,
					FeatureType.ATTRIBUTE);
		}
		if (getParameterAsBoolean(PARAM_INCLUDE_EDGE)) {
			ImagePlus implus = new ImagePlus("", imp.getImage());
			ImageProcessor ip = implus.getProcessor();
			ip.findEdges();

			double edge = ip.getHistogram()[255] / stats.area;
			feature.setFeature("Eginess", edge, Ontology.REAL,
					FeatureType.ATTRIBUTE);
		}
		imObj.addFeatures(feature);
		gsImOutput.deliver(imObj);
		globaFeature.deliver(feature);
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types
				.add(new ParameterTypeBoolean(
						PARAM_INCLUDE_MEAN,
						"Include mean value - average gray value in the image. This is the sum of the gray values of all the pixels in the image divided by the number of pixels.",
						true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_MEDIAN,
				"Include median value.", true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_KURTOSIS,
				"The fourth order moment about the mean.", true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_STDEV,
				"Standard deviation of the gray values used to generate the mean gray value.", true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_PEAK,
				"Include peak color from histogram and its relative value.", true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_MIGV,
				"Include Minimum Gray value in the image.", true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_MAGV,
				"Include Maximum Gray value in the image.", true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_MASSCENT,
				"This is the brightness-weighted average of the x and y coordinates all pixels in the image.", true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_SKEWNESS,
				"The third order moment about the mean.", true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_ARFRAC,
				"The percentage of non-zero pixels.", true));
		types.add(new ParameterTypeBoolean(PARAM_INCLUDE_EDGE,
				"Relative count of white (255) pixels after applying \"Find edges\" operator", true));
		return types;
	}
}
/*
 * http://imagejdocu.tudor.lu/doku.php?id=gui:analyze:set_measurements Area -
 * Area of selection in square pixels. Area is in calibrated units, such as
 * square millimeters, if Analyze>Set Scale was used to spatially calibrate the
 * image.
 * 
 * Mean Gray Value - Average gray value within the selection. This is the sum of
 * the gray values of all the pixels in the selection divided by the number of
 * pixels. Reported in calibrated units (e.g., optical density) if
 * Analyze>Calibrate was used to calibrate the image. For RGB images, the mean
 * is calulated by converting each pixel to grayscale using the formula
 * gray=0.299*red+0.587*green+0.114*blue if Weighted RGB Conversion is checked
 * in Edit>Options>Conversions. or the formula gray=(red+green+blue)/3 if not
 * checked.
 * 
 * Standard Deviation - Standard deviation of the gray values used to generate
 * the mean gray value.
 * 
 * Modal Gray Value - Most frequently occurring gray value within the selection.
 * Corresponds to the highest peak in the histogram.
 * 
 * Min & Max Gray Value - Minimum and maximum gray values within the selection.
 * 
 * Center of Mass - This is the brightness-weighted average of the x and y
 * coordinates all pixels in the selection.
 * 
 * Median - The median value of the pixels in the image or selection.
 * 
 * Skewness - The third order moment about the mean. The documentation for the
 * Moment Calculator plugin explains how to interpret spatial moments.
 * 
 * Kurtosis - The fourth order moment about the mean.
 * 
 * Area Fraction - The percentage of pixels in the image that have
 * been highlighted in red using Image/Adjust/Threshold. For non-thresholded
 * images, the percentage of non-zero pixels.
 */